package enchanting;

import java.util.Random;
import java.util.ArrayList;

/**
 ** Wrapper class for objects in lists or variables
 **/
public class EObject {
	
	public static enum ETYPE {
    	NUMBER,
    	TEXT,
    	BOOLEAN,
    	LIST,
    	COMMAND,
    	REPORTER,
    	PREDICATE
 	}
    
    // Member variables
  	protected Object data;
    protected ETYPE type;
    
	protected Object getValue()  { return data; } 
    
    // Returns the type of the object
	public synchronized ETYPE	getType() 			{ return type; } 
	
	public synchronized boolean isNumber() 			{ return type == ETYPE.NUMBER; }
 	public synchronized boolean isText()			{ return type == ETYPE.TEXT; }
 	public synchronized boolean isBoolean()			{ return type == ETYPE.BOOLEAN; }
 	public synchronized boolean isList()			{ return type == ETYPE.LIST; }
 	public synchronized boolean isCommand()			{ return type == ETYPE.COMMAND; }
 	public synchronized boolean isReporter()		{ return type == ETYPE.REPORTER; }
 	public synchronized boolean isPredicate()		{ return type == ETYPE.PREDICATE; }
	
	// likely a better place for this ... sigh.
	public static Random random = new Random();

	// Equality	
 	public synchronized boolean equals(EObject o) {
   		return this.type.equals(o.getType()) && 
   			this.data.equals(o.getValue());
   	}
  	
  	// Comparison
	public synchronized boolean greaterThan(EObject o, boolean reverse) {
			
		// always return false if types are different
		if (type != o.getType())
			return false;
				
		// check for equality first
		if (equals(o))
			return false;
			
		if (type == ETYPE.NUMBER)
			return asNumber() > o.asNumber() ^ reverse;
		if (type == ETYPE.TEXT)
			return asText().compareTo(o.asText()) > 0 ^ reverse;
		if (type == ETYPE.BOOLEAN)
			return asBoolean().compareTo(o.asBoolean()) > 0 ^ reverse;
				
		// return false for all other types
		return false;

	}

	// toString
	@Override
	public synchronized String toString() {
		return data.toString();
	}
	
	// for communication to Enchanting
	protected String toSmalltalkString() {
		//System.out.println(type);
		if (isList()) {
			String text = "#(";
			for (EObject e : asList()) { text += e.toSmalltalkString() + " "; }
			return text + ")";
		}
		if (isText()) {
			StringBuilder str = new StringBuilder();
			str.append('\'');
			for (char c : asText().toCharArray()) {
				if (c == '\'') { str.append('\''); } // escape single quote
				str.append(c);
			}
			str.append('\'');
			return str.toString();
		}
		// Return unquoted String for all other types (more to do here for cmd, pred, rep)
		return asText();
	}
	

	// Setters and Constructors
	//
	
	// UNDEFINED
	public EObject()					{ setValue(""); }
	
	// EXISTING EOBJECT
	public EObject(EObject value) 		{ setValue(value); }
	
	public synchronized void setValue(EObject value) {
		this.data = value.data;
		this.type = value.type;
	}
	
	// NUMBERS
	//
	public EObject(float value) 		{ setValue(value); }
	public EObject(int value) 			{ setValue((float)value); }
	public EObject(double value) 		{ setValue((float)value); }

	public synchronized void setValue(float value) {
		type = ETYPE.NUMBER;
		data = value;
	}
	
	public synchronized void setValue(int value)	{ setValue((float)value); }
	public synchronized void setValue(double value)	{ setValue((float)value); }

	
	// STRING
	//
	public EObject(String value) 		{ setValue(value); }

	public synchronized void setValue(String value)
	{
		// Assumptions:
		// the value passed in is in string format.
		// Commands, predicates, reporters and lists do not have string representations.
		
		// Numbers and booleans do have string representations.
		try {
			float f = Float.parseFloat(value);
			setValue(f);
			return;
		} catch (NumberFormatException e) {}
		if (value.equalsIgnoreCase("false")) {
			setValue(false);
			return;
		}
		if (value.equalsIgnoreCase("true")) {
			setValue(true);
			return;
		}

		// The value here is best represented by a string	
		type = ETYPE.TEXT;
		data = value;		
	}
	
	// BOOLEAN
	//
	public EObject(boolean value) 		{ setValue(value); }

	public synchronized void setValue(boolean value) {
		type = ETYPE.BOOLEAN;
		data = value;
	}	
	
	// LIST
	//
	public EObject(ArrayList<EObject> value) 	{ setValue(value); }
	
	public static EObject createList() {
		return new EObject(new ArrayList<EObject>());
	}
	
	public static EObject createList(EObject ... values) {
		EObject list = new EObject(new ArrayList<EObject>());
		for (int i = 0; i < values.length; i++) {
			list.addItemToList(values[i]);
		}
		return list;
	}

	public static EObject createList(float ... values) {
		EObject list = new EObject(new ArrayList<EObject>());
		for (int i = 0; i < values.length; i++) {
			list.addItemToList(values[i]);
		}
		return list;
	}
	
	public EObject createListCopy() {
		ArrayList<EObject> original = asList();
		int len = original.size();
		EObject newList = new EObject(new ArrayList<EObject>(len));
		for (int i = 0; i < len; i++) {
			newList.addItemToList(original.get(i));
		}
		return newList;
	}
	
	public static EObject createHSVListFromRGB(int r, int g, int b) {
		// Takes red, green, and blue values in range (0-255)
		// and returns hue (0-360), and saturation and value (0-100)
		// in a list

		// r, g, and b, as they come in, aren't necessarily in the right range!
		// [may be useful for calibration, but gives hsv values that fall outside of range].
		if (r > 255) { r = 255; } else if (r < 0) { r = 0; }
		if (g > 255) { g = 255; } else if (g < 0) { g = 0; }
		if (b > 255) { b = 255; } else if (b < 0) { b = 0; }

		float h, s, v;
		int max = Math.max(r, Math.max(g, b));
		int min = Math.min(r, Math.min(g, b));
		float span = (float)max - (float)min;
		
		// calculate hue
		if (span == 0.0f) {
			h = 0.0f;
		} else {
			if (r == max) {
				h = (float)(g - b) / span * 60.0f;
			} else if (g == max) {
				h = (float)(b - r) / span * 60.0f + 120.0f;
			} else {
				h = (float)(r - g) / span * 60.0f + 240.0f;
			}
			if (h < 0.0f) { h += 360.0f; }
		}
		
		// calculate saturation
		if (max == 0) {
			s = 0.0f;
		} else {
			s = 100.0f * span / (float)max;
		}
		
		// calculate value or brightness
		v = 100.0f * max / 255.0f; 	
		
		// return EObject.createList(h, s, v);	
		return EObject.createList((float)(Math.round(h)), (float)(Math.round(s)), (float)(Math.round(v)));	
	}

	public synchronized void setValue(ArrayList<EObject> value) {
	 	type = ETYPE.LIST;
	 	data = value;
	}
	
	// COMMANDS
	//
	public EObject(Command value)		{ setValue(value); }
	
	public void setValue(Command value) {
		type = ETYPE.COMMAND;
		data = value;
	}
	
	// REPORTER
	//
	public EObject(Reporter value)		{ setValue(value); }

	public synchronized void setValue(Reporter value) {
		type = ETYPE.REPORTER;
		data = value;
	}

	// PREDICATE
	//
	public EObject(Predicate value)		{ setValue(value); }

	public synchronized void setValue(Predicate value) {
		type = ETYPE.PREDICATE;
		data = value;
	}

	// OBJECTS
	//
	@SuppressWarnings("unchecked")
	public EObject(Object obj) { 
		if (obj instanceof Command)
			setValue((Command)obj);
		else if (obj instanceof Reporter)
			setValue((Reporter) obj);
		else if (obj instanceof Predicate)
			setValue((Predicate) obj);
		else if (obj instanceof ArrayList)
			setValue((ArrayList<EObject>)obj);
		else
			setValue(obj.toString()); 
	}
	
	// GETTERS
	//
	public synchronized String asText() { 
		return data.toString(); 
	}
	
	public synchronized Float asNumber() { 
		// Assumption: value is already represented by best type
		if (type == ETYPE.NUMBER) {
			return (Float)data;
		}
		if (type == ETYPE.BOOLEAN) {
			return (Boolean)data ? 1.0f : 0.0f;
		}
		
		// The value can not be represented as a number
		return 0.0f;
	}
	
	public synchronized int asInteger() {
		return asNumber().intValue();
	}
	
	public synchronized Boolean asBoolean() { 
		// Assumption: value is already represented by the best type
		if (type == ETYPE.BOOLEAN) {
			return (Boolean)data;
		}
		if (type == ETYPE.NUMBER) {
			// do we need to check the absolute against a really tiny value instead?
			return (Float)data != 0.0f;
		}
		if (type == ETYPE.TEXT) {
			if ((String)data == "") return false;
			if (((String)(data)).equalsIgnoreCase("false")) return false;
		}
		
		// the value is either true, or not a boolean
		return true;
	}
	
	@SuppressWarnings("unchecked")
	public synchronized ArrayList<EObject> asList() { 
		try {
			return (ArrayList<EObject>)data;
		} catch (ClassCastException ex) {
			return new ArrayList<EObject>();
		}
	}
				
/**	public synchronized Command asCommand() { 
		try {
			return (Command)data;
		} catch (ClassCastException ex) {
			return new Command(); 
		}
	}
	
	public synchronized Reporter asReporter() { 
		try {
			return (Reporter)data;
		} catch (ClassCastException ex) {
			return new Reporter(); 
		}
	}
	
	public synchronized Predicate asPredicate() { 
		try {
			return (Predicate)data;
		} catch (ClassCastException ex) {
			return new Predicate(); 
		}
	}*/
	public synchronized CustomBlock asCustomBlock() {
		try {
			return (CustomBlock)data;
		} catch (ClassCastException ex) {
			return new Command(); //Could return either cmd / pred / reporter here
		}
	}
	
	// SPECIAL FUNCTIONS
	public synchronized void increment(float inc) {
		// returns true if the value changed
		// assumption: we are already set as the best type
		if (type == ETYPE.NUMBER) {
			if (inc == 0) { return; }
			float result = (Float)data + inc;
			data = result;
		} else {
			setValue(asNumber() + inc);
		}
		sendUpdate();
	}
	
	// LIST FUNCTIONS  ----- TODO integer flags (-2,-3) for any / last special cases?
	//
	
	// List index
	private synchronized int listIndex(EObject index) {
		if (index.isText()) {
			return listIndex(index.asText());
		} else {
			return listIndex(index.asInteger());
		}
	}
	
	private synchronized int listIndex(String index) {
    	// returns -1 if the index is not valid
    	// note that in Enchanting used 1-based indices, and java uses 0-based indices
    	if (!isList()) return -1;
    	int length = asList().size();
    	if (length < 1) return -1;
    	
    	if (index.equals("any")) {
    		return random.nextInt(length);
    	} else if (index.equals("last")) {
    		return length - 1;
    	}
    	
    	//Attempt to parse as number
    	try { 
    		int indx = Integer.parseInt(index);
    		return listIndex(indx);
    	} catch (NumberFormatException ex) {
    		return -1;
    	}
    }
    private synchronized int listIndex(int index) {
    	if (index > 0 && index <= asList().size())
    		return index - 1;
    	return -1;
    }

    // Designed only to check the output of 'listIndex'
    private synchronized boolean validListIndex(int index) {
    	return index > -1;
    }
    
    // List element adding
    public synchronized void addItemToList(EObject item) {
    	if (isList()) {
    		asList().add(new EObject(item));
    		sendUpdate();	
    	}
    }
    public synchronized void addItemToList(Object item) {
    	addItemToList(new EObject(item));
    }
    
    // List element inserting
	public synchronized void insertItemInList(EObject index, EObject item) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().add(indx, item);
    		sendUpdate();
    	}
    }
    public synchronized void insertItemInList(EObject index, Object item) {
    	insertItemInList(index, new EObject(item));
    }
    

	// List element getting
	public synchronized Object getItemFromList(EObject index) {
		int indx = listIndex(index);
	    if (validListIndex(indx))
    		return asList().get(indx);
    	return "";
    }
   
    // List element setting - where possible use the existing EObject rather than creating a new one
    public synchronized void setItemInList(EObject index, EObject newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().set(indx, newValue);
    		sendUpdate();
    	}	
    }
    public synchronized void setItemInList(EObject index, float newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }
    public synchronized void setItemInList(EObject index, int newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    	}	
    }
    public synchronized void setItemInList(EObject index, double newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }
    public synchronized void setItemInList(EObject index, String newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }
    public synchronized void setItemInList(EObject index, boolean newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }  
    public synchronized void setItemInList(EObject index, ArrayList<EObject> newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }
    public synchronized void setItemInList(EObject index, Command newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }
  	public synchronized void setItemInList(EObject index, Reporter newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }
  	public synchronized void setItemInList(EObject index, Predicate newValue) {
    	int indx = listIndex(index);
    	if (validListIndex(indx)) {
    		asList().get(indx).setValue(newValue);
    		sendUpdate();
    	}	
    }
    
	// List element removing
    public synchronized void removeItemFromList(EObject index) {
    	if (index.isText() && index.asText().equals("all")) {
			if (isList()) {
				asList().clear();
				sendUpdate();
			}
    	} else {
	    	int indx = listIndex(index);   
    		if (validListIndex(indx)) {
    			asList().remove(indx);
    			sendUpdate();
    		}
    	}
    }

	// Other code has been told this EObject is a color, an HSV list, and would like it
	public synchronized HSV asHSV() {
		HSV color = new HSV();	// invalid as initially created
		if (isList()) {
			ArrayList<EObject> list = this.asList();
			if (list.size() >= 3) {
				color.hue = list.get(0).asNumber();
				color.sat = list.get(1).asNumber();
				color.val = list.get(2).asNumber();			
			}
		}
		return color;
	}
    
    // Contains
    public synchronized boolean contains(Object o) {
    	if (isList()) {
    		for (EObject e : asList()) {
    			if (e.equals(o))
    			return true;
    		}
    	}
    	return false;
    }
    
    // Length
    public synchronized int length() {
    	if (isList())
    		return asList().size();
    	else
    		return 0;
    }
    
    // Update method overridden by variable
    protected void sendUpdate() {
    }
}